-- Given a an integer representing a date (interpreted as an epoch timestamp),
-- returns the name of a new messages table partition.
CREATE OR REPLACE FUNCTION message_table_name( tbl_date INTEGER )
RETURNS CHAR(30) AS $$
DECLARE
the_hour CHAR(2);
BEGIN
	the_hour := (select extract(hour from (select TIMESTAMP 'epoch' at time zone 'UTC' + tbl_date * INTERVAL '1 second')::timestamp))::text;
	IF (the_hour::int < 10) THEN
		the_hour := '0'||the_hour::text;
	END IF;

	RETURN 'messages_' || REPLACE(((select TIMESTAMP 'epoch' at time zone 'UTC' + tbl_date * INTERVAL '1 second')::DATE)::text, '-','_') || '_' || the_hour::text;
END;
$$
LANGUAGE plpgsql;


-- Maintains the partitioning scheme by dropping tables that are older
-- than 'keep_hours'+1 hours. Returns the number of dropped tables.
-- We add one because the keep_hours setting acts as a lower bound (as opposed
-- to an upper bound) for queryable data.
CREATE OR REPLACE FUNCTION partition_maintenance_function()
RETURNS INTEGER AS $$
DECLARE
_count INTEGER;
_keep_hours INTEGER;
_lower_bound TIMESTAMP;
_result RECORD;
BEGIN
	_keep_hours := (SELECT keep_in_db+1 FROM settings);
	_lower_bound := (localtimestamp - ( _keep_hours || ' hours')::interval)::timestamp;
	_count = 0;
	FOR _result IN (SELECT table_name from information_schema.tables
		WHERE table_catalog = 'op5lsu'
		AND table_name LIKE 'message_%') LOOP
		IF  char_length(substring(_result.table_name from '[0-9_]*$')) <> 0 AND
			(_lower_bound > to_timestamp(substring(_result.table_name from '[0-9_]*$'), '_YYYY_MM_DD_HH24')) THEN
			EXECUTE 'DROP TABLE ' || quote_ident(_result.table_name);
			_count = _count + 1;
		END IF;
	END LOOP;
	RETURN _count;
END
$$
LANGUAGE plpgsql;

-- Creates a new partition of the messages table, based on the `tbl_date`
-- parameter, if the corresponding table does not exist. The name of the new
-- table is determined by a call to message_table_name().
-- A partition in this case is a table that inherits from the base messages
-- table, but only stores messages for one whole hour. Calling this function
-- also triggers a partition maintenance round.
--
-- Returns the name of the table, regardless of whether it was just created or
-- not.
CREATE OR REPLACE FUNCTION create_table_function( tbl_date INTEGER )
RETURNS CHAR(30) AS $$
DECLARE
table_name CHAR(30);
_hour CHAR(2);
_starttime integer;
_endtime integer;
BEGIN
	table_name := message_table_name(tbl_date);
	IF EXISTS (
		SELECT *
		FROM   pg_catalog.pg_tables
		WHERE    tablename  = table_name
		) THEN
		RETURN table_name;
	ELSE
		_hour := (select extract(hour from (select TIMESTAMP 'epoch' at time zone 'UTC' + tbl_date * INTERVAL '1 second')::timestamp))::text;
		_starttime := (select EXTRACT(EPOCH FROM (select (select TIMESTAMP 'epoch' at time zone 'UTC' + tbl_date * INTERVAL '1 second')::date::text || ' ' || _hour || ':00' )::timestamp));
		_endtime := (select _starttime + 3600);
		EXECUTE ( 'CREATE TABLE '
			|| quote_ident(table_name)
			|| ' ( CHECK ( rtime < '
			|| _endtime
			|| ' AND rtime >= '
			|| _starttime
			|| '), PRIMARY KEY (id, rtime)
		) INHERITS (messages);' );
		PERFORM partition_maintenance_function();
	END IF;
	RETURN table_name;
END;
$$
LANGUAGE plpgsql;


-- This trigger is run on inserts to the messages table, and takes care of
-- creating new partitions as well as redirecting messages into the correct
-- table.
CREATE OR REPLACE FUNCTION messages_insert_trigger()
RETURNS TRIGGER AS $$
DECLARE
table_name CHAR(30);
BEGIN
	table_name := create_table_function(NEW.rtime);
	EXECUTE 'INSERT INTO ' ||  quote_ident(table_name) || ' (ip, rtime, facility, severity, mtime, host, ident, pid, event, msg, fts) VALUES($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11);'
		USING NEW.ip, NEW.rtime, NEW.facility, NEW.severity, NEW.mtime, NEW.host, NEW.ident, NEW.pid, NEW.event, NEW.msg, NEW.fts;
	RETURN NULL;
END;
$$
LANGUAGE plpgsql;
